package com.mysterion.xml;

/**
 * XMLBoss version yeri.1.0
 * FileStream close() included!
 */

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.StringBufferInputStream;
import java.io.UnsupportedEncodingException;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.SortedSet;
import java.util.TreeMap;
import java.util.TreeSet;
import java.util.Vector;
import oracle.xml.parser.v2.XMLDOMImplementation;
import oracle.xml.parser.v2.XMLDocument;
import oracle.xml.parser.v2.XSLProcessor;
import oracle.xml.parser.v2.XSLStylesheet;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import java.util.regex.Pattern; 
import java.util.regex.Matcher;

import org.w3c.dom.ls.DOMImplementationLS;
import org.w3c.dom.ls.LSInput;
import org.w3c.dom.ls.LSParser;

/* Included from JAXP for Xpath Expression Evaluation */
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;

/*
 *  This class was created to reduce the unwarranted complexity involved with 
 *  trying to build and parse XML and XSLT files in Java.  It provides five
 *  methods, four for creating XML and XSLT files, and one for transforming an
 *  XML document using an XSLT stylesheet:
 *  
 *  XML methods:
 *      public XMLDocument XMLFromFile (String filePath)
 *          - parse a file and return an XML document
 *      public XMLDocument XMLFromString (String xmlInput)
 *          - parse a string and return an XML document
 *      
 *  XSLT methods:
 *      public XSLStylesheet XSLFromFile (String filePath)
 *          - parse a file and return an XSLT stylesheet
 *      public XSLStylesheet XSLFromString (String xslInput)
 *          - parse a string and return an XSLT stylesheet
 *      
 *  Transformation:
 *      public String Transform (XMLDocument xDoc, XSLStylesheet xslDoc)
 *          - return the result of applying a stylesheet to an xml document
 * 
 *  Filtering methods:      
 *    public NodeList SelectNodes (Node xNode, String xPathExpression)
 *          - return the nodes selected by the given XPath Expression
 *            (NOTE: 'xNode' can be an XML document or any sub-node)
 *    public Node SelectNode (Node xNode, String xPathExpression)
 *          - return the first node selected by the given XPath Expression
 *            (NOTE: 'xNode' can be an XML document or any sub-node)
 *    public Vector ChildNodes (Node xNode)
 *          - return a vector which contains ONLY the ELEMENT_NODE children of the given node
 *    public Vector ChildNodes (Node xNode, String tagName)
 *          - return a vector which contains ONLY the ELEMENT_NODE children of the given node
 *            which have the given tag name
 *    
 *  Helper methods:
 *      public SortedSet AttributeNames (Node xmlNode)
 *          - return an alphabetical list of the attributes of a given node
 *      public TreeMap Attributes (Node xmlNode)
 *          - return a list of key/value pairs for the attributes of a given node
 *          
 *      public String TextContent (Node xmlNode)
 *          - return the text content of the given node
 *      public String AsText (Node xmlNode)
 *          - return the node & its descendants as an XML string
 *      public XMLDocument AsXML (Node xmlTree)
 *          - return the node & its descendants as a new XML document      
 */

public class XMLBoss
{
  static XPathFactory xpf = XPathFactory.newInstance();

  /* --------------------------------------------------------------------------------- */
  /* ----------------------------- FILTER METHODS ------------------------------------ */  
  /* --------------------------------------------------------------------------------- */

  static public Vector ChildNodes (Node xNode)
  /*
   * Given an XML element, return only the child nodes which are
   * other XML elements (no text, cdata, comment, etc. nodes)
   */
  {
    Vector result = new Vector();
    NodeList children = xNode.getChildNodes();

    for (int i = 0; i < children.getLength(); i++)
      {
        if (children.item(i).getNodeType() == Node.ELEMENT_NODE)
          {
            result.add (children.item(i));
          }
      }

    return result;
  }

  static public Vector ChildNodes (Node xNode, String tagName)
  /*
   * Given an XML element, return only the child nodes which are
   *  elements and whose tag names equal the passed-in string
   */
  {
    Vector result = new Vector();
    NodeList children = xNode.getChildNodes();

    for (int i = 0; i < children.getLength(); i++)
      {
        if (children.item(i).getNodeType() == Node.ELEMENT_NODE)
          {
            org.w3c.dom.Element eTemp = (org.w3c.dom.Element)children.item(i);
            if (tagName.equalsIgnoreCase(eTemp.getTagName()))
              {
                result.add(eTemp);
              }
          }
      }

    return result;
  }

  static public NodeList SelectNodes (Node xNode, String xPathExpression) throws Exception
  /* Given an XPath expression and an XML element, return a list of nodes which are
   * selected by the expressions
   * 
   * INPUTS:
   *    xDoc : the XML document object
   *    xPathExpression : the Xpath Expression to run against the document
   * 
   * OUTPUT:
   *    a Nodelist containing the nodes selected by the expression
   */
  {
    Node result = null;
    NodeList nList = null;

    XPath xpath = xpf.newXPath();		
    xpath.compile(xPathExpression);

    if (xNode.getNodeType() == Node.DOCUMENT_NODE)
      {
        nList = (NodeList)xpath.evaluate(xPathExpression, xNode, XPathConstants.NODESET);
      }
    else
      {
        /* Create a new XML document from the node which was passed to us
         * NOTE: this is necessary because using a "//" expression on a node
         *       forces the XPath expression evaluator to go back and look at
         *       the entire XML document, not just the descendants of the node
         */
        XMLDocument newDoc = XMLBoss.AsXML(xNode);
        if (newDoc != null)
          {
            nList = (NodeList)xpath.evaluate(xPathExpression, newDoc, XPathConstants.NODESET);
          }
      }

    return nList;
  }

  static public Node SelectNode (Node xNode, String xPathExpression) throws Exception
  /* Given an XPath expression and an XML element, return a the first node which
   * is selected by the expression
   * 
   * INPUTS:
   *    xDoc : the XML document object
   *    xPathExpression : the Xpath Expression to run against the document
   * 
   * OUTPUT:
   *    the first node selected by the expression
   */
  {
    Node result = null;
    NodeList nList = XMLBoss.SelectNodes (xNode, xPathExpression);

    if (nList.getLength() > 0)
      {result = nList.item(0);}

    return result;
  }

  /* --------------------------------------------------------------------------------- */
  /* ----------------------------- HELPER METHODS ------------------------------------ */  
  /* --------------------------------------------------------------------------------- */

  static public String TextContent (Node xmlNode, boolean stripWhiteSpace)
  {
    String result = "";
    StringBuffer  sTemp = new StringBuffer();

    NodeList children = xmlNode.getChildNodes();
    for (int i = 0; i < children.getLength(); i++)
      {
        Node iTemp = children.item(i);
        if (iTemp.getNodeType() == Node.TEXT_NODE)
          {
            sTemp.append (iTemp.getNodeValue());
          }
        if (iTemp.getNodeType() == Node.CDATA_SECTION_NODE)
          {
            sTemp.append (iTemp.getNodeValue());
          }
      }

    result = sTemp.toString();
    if (stripWhiteSpace)
      {
        Pattern p = Pattern.compile("\\s+");
        Matcher m = p.matcher(result); // get a matcher object
        result = m.replaceAll(" ");
      }
    return result.trim();
  }

  static public String TextContent (Node xmlNode)
  /*
   * Given a node, return a concatenation of all of the text nodes which are its
   * immediate children
   */
  {
    return XMLBoss.TextContent (xmlNode, true);
  }


  static public String AsText (Node xmlNode) throws UnsupportedEncodingException
  /* 
   * Given a valid XML node, return a serialized version of the node
   * and all of its descendants
   */
  {        
    return DOMWriter.print (xmlNode, false);
  }

  static public XMLDocument AsXML (Node xmlNode) throws UnsupportedEncodingException
  /* 
   * Given a valid XML node, return a new XML Document which consists of that
   * node and its descendants
   */
  {        
    return XMLBoss.XMLFromString(XMLBoss.AsText(xmlNode));
  }

  static public SortedSet AttributeNames (Node xmlNode)
  /*
   * Given a node, return a sorted list of it's attribute names   * 
   */
  {
    TreeSet result = new TreeSet();
    try
      {
        NamedNodeMap aList = xmlNode.getAttributes();

       for (int i = 0; i < aList.getLength(); i++)
          {
            Node attr = aList.item(i);
            result.add(attr.getNodeName());         
          }
      }
    catch (Exception e) {}
    return result;
  }

  static public TreeMap Attributes (Node xmlNode)
 /*
   * Given a node, return a hashtable of it's attribute names and values  * 
   */
  {
    TreeMap result = new TreeMap();

    try
      {
        NamedNodeMap aList = xmlNode.getAttributes();

        for (int i = 0; i < aList.getLength(); i++)
          {
            Node attr = aList.item(i);
            result.put(attr.getNodeName(), attr.getNodeValue());        
          }
      }
    catch (Exception e)
      {

      }
    return result;
  }

  /* --------------------------------------------------------------------------------- */
  /* ----------------------------------- XML ----------------------------------------- */  
  /* --------------------------------------------------------------------------------- */

  static public XMLDocument XMLFromFile (File file)
  throws FileNotFoundException, IOException
  {
    FileInputStream sTemp = new FileInputStream(file);               
    DOMImplementationLS impl = new XMLDOMImplementation();
    LSInput lsInput= impl.createLSInput();
    lsInput.setByteStream(sTemp);
    LSParser parser = impl.createLSParser(DOMImplementationLS.MODE_SYNCHRONOUS, null);
    XMLDocument document = (XMLDocument)parser.parse(lsInput);  
    sTemp.close();
    return document;
  }
  
  static public XMLDocument XMLFromInputStream (InputStream inStream)
	{                 
	    DOMImplementationLS impl = new XMLDOMImplementation();
	    LSInput lsInput= impl.createLSInput();
	    lsInput.setByteStream(inStream);
	    LSParser parser = impl.createLSParser(DOMImplementationLS.MODE_SYNCHRONOUS, null);
	    XMLDocument document = (XMLDocument)parser.parse(lsInput); 
	    return document;
	}

  static public XMLDocument XMLFromString (String xmlInput)
  {                 
    DOMImplementationLS impl = new XMLDOMImplementation(); 
    LSInput lsInput= impl.createLSInput();
    lsInput.setStringData(xmlInput);
    LSParser parser = impl.createLSParser(DOMImplementationLS.MODE_SYNCHRONOUS, null);
    XMLDocument document = (XMLDocument)parser.parse(lsInput);
    return document;
  }

  /* --------------------------------------------------------------------------------- */
  /* ----------------------------------- XSLT ---------------------------------------- */  
  /* --------------------------------------------------------------------------------- */

  static public XSLStylesheet XSLFromFile (File file) throws java.io.FileNotFoundException, oracle.xml.parser.v2.XSLException
  {
    // Create an XSL Processing object using the input file
    FileInputStream sTemp = new FileInputStream(file);
    XSLProcessor xslp = new XSLProcessor();
    XSLStylesheet xs = xslp.newXSLStylesheet((InputStream)sTemp);
    return xs;
  }

  /**
   * @deprecated
   * @param xslInput
   * @return
   * @throws java.io.FileNotFoundException
   * @throws oracle.xml.parser.v2.XSLException
   */
  static public XSLStylesheet XSLFromString (String xslInput) throws java.io.FileNotFoundException, oracle.xml.parser.v2.XSLException
  {
    // Create an XSL Processing object using the input stream
      StringBufferInputStream sTemp = new StringBufferInputStream(xslInput);
      XSLProcessor xslp = new XSLProcessor();
      XSLStylesheet xs = xslp.newXSLStylesheet((InputStream)sTemp);
      return xs;
  }

  static public String Transform (XMLDocument xDoc, XSLStylesheet xslDoc) throws oracle.xml.parser.v2.XSLException, java.io.IOException
  {
      XSLProcessor xslp = new XSLProcessor();
      // Create an output stream into which the results will be placed
      OutputStream bs = new ByteArrayOutputStream ();
      // Invoke the XSL Processor
      xslp.processXSL(xslDoc, xDoc, (OutputStream)bs);
      // Return the results
      return bs.toString();
  }

}
