/**
 * 
 */
package edu.cmu.taghelpertools.middle_layer;

import java.util.ArrayList;
import java.util.Iterator;

import edu.cmu.taghelpertools.tokenizer.EnglishTokenizer;
import edu.cmu.taghelpertools.tokenizer.IndexedFeature;
import edu.cmu.taghelpertools.tokenizer.StopwordManager;
import edu.cmu.taghelpertools.tokenizer.Token;
import edu.cmu.taghelpertools.tokenizer.Tokenizer;
import edu.cmu.taghelpertools.tokenizer.Tokens;
import edu.cmu.taghelpertools.util.HelperFunctions;
import gnu.trove.THashMap;
import gnu.trove.THashSet;

/**
 * FeatureSelector does all the feature selection. In other words, it is told
 * which features to extract from the text (e.g., unigrams, bigrams, POS
 * bigrams, sentence length, punctuation, etc), and extracts them. It is also
 * responsible for creating the "DICT" file, which is a flat-file that contains
 * all the options used and all the features for a classifier.
 * 
 * @author Jaime Arguello, Hao-Chuan Wang, Yi-Chia Wang
 * 
 */
public class FeatureSelector {

	public static final String CONTAINS_NON_STOPWORD = "CONTAINS_NON_STOP";
	public static final String LINE_LENGTH = "LINE_LENGTH";
	private static THashMap punctStringMap;

	private static String COMMA = "COMMA";
	private static String PERIOD = "PERIOD";
	private static String SEMI_COLON = "SEMI_COLON";
	private static String COLON = "COLON";
	private static String DASH = "DASH";
	private static String QUESTION_MARK = "QUESTION_MARK";
	private static String EXCLAMATION_MARK = "EXCLAMATION_MARK";
	private static String ASTERISK = "ASTERISK";
	private static String AMPERSAND = "AMPERSAND";
	private static String AT_SIGN = "AT_SIGN";
	private static String NUMBER_SIGN = "NUMBER_SIGN";
	private static String PERCENT_SIGN = "PERCENT_SIGN";
	private static String OPEN_PAREN = "OPEN_PAREN";
	private static String CLOSE_PAREN = "CLOSE_PAREN";
	private static String OPEN_CURLY = "OPEN_CURLY";
	private static String CLOSE_CURLY = "CLOSE_CURLY";
	private static String LESS_THAN = "LESS_THAN";
	private static String GREATER_THAN = "GREATER_THAN";
	private static String DOUBLE_QUOTE = "DOUBLE_QUOTE";
	private static String DOLLAR ="DOLLAR";
	private static String BACKSLASH = "BACKSLASH";
	private static String FORWARDSLASH = "FORWARDSLASH";
	private static String PLUS_SIGN = "PLUSSIGN";
	private static String TILDE = "TILDE";
	private static String CARAT = "CARAT";
	private static String LEFT_SQUARE_PAREN = "LEFTSQUAREPAREN";
	private static String RIGHT_SQUARE_PAREN = "RIGHTSQUAREPAREN";

	static {
		punctStringMap = new THashMap();
		punctStringMap.put("<COMMA>", FeatureSelector.COMMA);
		punctStringMap.put("<PERIOD>", FeatureSelector.PERIOD);
		punctStringMap.put("<SEMICOLON>", FeatureSelector.SEMI_COLON);
		punctStringMap.put("<COLON>", FeatureSelector.COLON);
		punctStringMap.put("<DASH>", FeatureSelector.DASH);
		punctStringMap.put("<QUESTIONMARK>", FeatureSelector.QUESTION_MARK);
		punctStringMap.put("<EXCLAMATIONMARK>", FeatureSelector.EXCLAMATION_MARK);
		punctStringMap.put("<ASTERISK>", FeatureSelector.ASTERISK);
		punctStringMap.put("<AMPERSAND>", FeatureSelector.AMPERSAND);
		punctStringMap.put("<ATSIGN>", FeatureSelector.AT_SIGN);
		punctStringMap.put("<NUMBERSIGN>", FeatureSelector.NUMBER_SIGN);
		punctStringMap.put("<PERCENTSIGN>", FeatureSelector.PERCENT_SIGN);
		punctStringMap.put("<OPENPAREN>", FeatureSelector.OPEN_PAREN);
		punctStringMap.put("<CLOSEPAREN>", FeatureSelector.CLOSE_PAREN);
		punctStringMap.put("<OPENCURLY>", FeatureSelector.OPEN_CURLY);
		punctStringMap.put("<CLOSECURLY>", FeatureSelector.CLOSE_CURLY);
		punctStringMap.put("<LESSTHAN>", FeatureSelector.LESS_THAN);
		punctStringMap.put("<GREATERTHAN>", FeatureSelector.GREATER_THAN);
		punctStringMap.put("<DOUBLEQUOTE>", FeatureSelector.DOUBLE_QUOTE);
		punctStringMap.put("<DOLLAR>", FeatureSelector.DOLLAR);
		punctStringMap.put("<BACKSLASH>", FeatureSelector.BACKSLASH);
		punctStringMap.put("<FORWARDSLASH>", FeatureSelector.FORWARDSLASH);
		punctStringMap.put("<PLUSSIGN>", FeatureSelector.PLUS_SIGN);
		punctStringMap.put("<TILDE>", FeatureSelector.TILDE);
		punctStringMap.put("<CARAT>", FeatureSelector.CARAT);
		punctStringMap.put("<LEFTSQUAREPAREN>", FeatureSelector.LEFT_SQUARE_PAREN);
		punctStringMap.put("<RIGHTSQUAREPAREN>", FeatureSelector.RIGHT_SQUARE_PAREN);

	}


	/**
	 * Contains the feature-selection settings
	 */
	private OptionsModel optionsModel;

	private static String FEAT_PREFIX_UNI = "";
	private static String FEAT_PREFIX_BI = "";
	private static String FEAT_PREFIX_POSBI = "";

	/**
	 * Contains the minimum sentence length. This is used because a sentence's
	 * length (an optional feature) is represented as: (X - minCount)/(maxCount -
	 * minCount), where X is the length of *this* sentence.
	 */
	private int minCount = Integer.MAX_VALUE;

	/**
	 * Contains the maximum sentence length. This is used because a sentence's
	 * length (an optional feature) is represented as: (X - minCount)/(maxCount -
	 * minCount), where X is the length of *this* sentence.
	 */
	private int maxCount = Integer.MIN_VALUE;

	/**
	 * Contains the union feature set.
	 */
	private THashMap featureMap;

	/**
	 * Sets up the FeatureSelector
	 * 
	 * @param optionsModel --
	 *            contains feature selection options
	 */
	public FeatureSelector(OptionsModel optionsModel)
	{
		try{
			this.optionsModel = optionsModel;
			this.featureMap = new THashMap();			
		}catch(Exception e){
			e.printStackTrace();
		}
	}



	/**
	 * This method is used when no advanced features are specified. It
	 * checks the static variables of the FeatureSelector, which are, in turn,
	 * set by the OptionsModel of the classifier being trained. The only feature
	 * that is not extracted in this method is sentence length. This method
	 * extracts all the features that have a string value, and sentence length
	 * has a numeric value.
	 * 
	 * @param text -
	 *            The text of the current DataPoint for which features are being
	 *            extracted.
	 * @return - The set of features extracted from this text.
	 */
	public THashMap getFeatureSet(String text) {
		return this.getFeatureSet(text, "");
	}

	public void getNGram(Tokens tokens, String text, String customPrefix, THashMap dataPointFeatureMap, int n){
		
		Iterator tokenIter = tokens.getIterator();
		int lim = tokens.getNumTokens();
		int cur = 0;
		Token[] tokensArr = new Token[lim];
		while (tokenIter.hasNext()) {
			tokensArr[cur++] = (Token)tokenIter.next();
		}

		for(int i = 1-n; i < tokensArr.length; i++){
			String featName = getFeatureName(n, tokensArr, i);
			if(featName != null && i>=0){
				addFeatToFeatureMap(dataPointFeatureMap, customPrefix+featName, tokensArr[i].getIndex(), featName.length());
			}
		}
	}
	
	public void getFeaturesFromPos(Tokens pos, Tokens surf, String customPrefix, THashMap dataPointFeatureMap){
		Iterator posItr = pos.getIterator();
		Iterator surfItr = surf.getIterator();
		ArrayList<Token> posTokens = new ArrayList<Token>();
		while(posItr.hasNext()) posTokens.add((Token)posItr.next());
		ArrayList<Token> surfTokens = new ArrayList<Token>();
		while(surfItr.hasNext()) surfTokens.add((Token)surfItr.next());
		System.out.println(posTokens.size() + " pos, " + surfTokens.size() + " surf.");
		String prevSurf = "";
		Token surfToken = null;
		int s = 0;
		for(int i = 0; i < posTokens.size(); i++){
			String curPosInd = posTokens.get(i).getIndexToken();
			String curPosSurf = posTokens.get(i).getSurfaceToken();
			String curSurf = "";
			if(s < surfTokens.size()){
				surfToken = surfTokens.get(s);				
				curSurf = (surfToken==null?"":surfToken.getSurfaceToken());
			}
			
			if(curPosInd.contains("EOL")){
				addFeatToFeatureMap(dataPointFeatureMap, FeatureSelector.FEAT_PREFIX_POSBI + customPrefix+curPosInd, surfToken.getIndex(), surfToken.getSurfaceToken().length());				
				continue;
			}
			
			String indTest = curPosInd.substring(curPosInd.length()-1, curPosInd.length());
			String surfTest = curPosSurf.substring(curPosSurf.length()-1, curPosSurf.length());
			boolean posPunctEnd = (indTest.matches("\\W+") || indTest.equals("_") || surfTest.matches("\\W+") || surfTest.equals("_")) && !(curPosInd.endsWith("$") && !curPosInd.endsWith("_$"));
			boolean surfPunctEnd = punctStringMap.contains(curSurf.toUpperCase());
			
			while(posPunctEnd != surfPunctEnd){
				if(posPunctEnd && !surfPunctEnd){
					String startInd = curPosInd.split("_")[0];
					String startSurf = curPosSurf.split("_")[0];
					i++;
					System.out.println("Skipping (1)(" + curPosInd + ", " + curPosSurf + ") which would be " + prevSurf + curSurf);
					curPosInd = posTokens.get(i).getIndexToken();
					curPosSurf = posTokens.get(i).getSurfaceToken();
					String[] indArr = curPosInd.split("_");
					String[] surfArr = curPosSurf.split("_");
					String endInd = indArr[indArr.length-1];
					String endSurf = surfArr[surfArr.length-1];
					curPosInd = startInd+"_"+endInd;
					curPosSurf = startSurf+"_"+endSurf;	
				}else if(!posPunctEnd && surfPunctEnd){
					System.out.println("Skipping (2)(" + curPosInd + ", " + curPosSurf + ") which would be " + prevSurf + curSurf);
					surfToken = surfTokens.get(++s);
					curSurf = surfToken.getSurfaceToken();
				}
				
				indTest = curPosInd.substring(curPosInd.length()-1, curPosInd.length());
				surfTest = curPosSurf.substring(curPosSurf.length()-1, curPosSurf.length());
				posPunctEnd = (indTest.matches("\\W+") || indTest.equals("_") || surfTest.matches("\\W+") || surfTest.equals("_")) && !(curPosInd.endsWith("$") && !curPosInd.endsWith("_$"));
				surfPunctEnd = punctStringMap.contains(curSurf.toUpperCase());
			}
			String surfStr = prevSurf+curSurf;
			
			addFeatToFeatureMap(dataPointFeatureMap, FeatureSelector.FEAT_PREFIX_POSBI + customPrefix+curPosInd, surfToken.getIndex(), surfStr.length());			
			
			String testSurf = surfStr.replaceAll("[^A-Za-z]", "").toLowerCase();
			String testPos = curPosSurf.replaceAll("[^A-Za-z]", "").toLowerCase();
			if(testSurf.equals(testPos) || !testSurf.startsWith(testPos) || testPos.length() == 0 || !curPosSurf.contains("_")){
				s++;
				prevSurf = curSurf+"_";
			}else{
				System.out.println("?????????????????????????????????????????????????????????????????????????????????????????????????????????");
			}
			System.out.println("Association " + i + ": " + curPosInd + " (" + curPosSurf + ", " + testPos + "), (" + surfStr + ", " + testSurf + ")");
		}
	}

	private String getFeatureName(int n, Token[] tokensArr, int i) {
		boolean containsNon = false;
		String featName = "";
		for(int j = 0; j < n; j++){
			if (optionsModel.isRemoveStopwords() && i+j >= 0 && i+j < tokensArr.length){
				String term = tokensArr[i+j].getSurfaceToken().trim().toLowerCase().split("([\\W|_&&[^'-<>]])")[0];
				if (StopwordManager.getInstance(EnglishTokenizer.STOPWORD_FILENAME).isStopword(term) || (FeatureSelector.punctStringMap.containsKey(term.toUpperCase()))) {
					if(!optionsModel.isStopwordsAll()){
						return null;
					}
				}else{
					containsNon = true;
				}
			}
			if(((n > 1 && (i+j > 0) && (i+j < tokensArr.length) && !optionsModel.isPunctuation()) || n == 1) && punctStringMap.contains(tokensArr[i+j].getIndexToken().toUpperCase())){ return null; }
			if(i+j < 0){
				featName += EnglishTokenizer.BOL;
			}else if(i+j < tokensArr.length){
				featName += tokensArr[i+j].getIndexToken();
			}else{
				featName += EnglishTokenizer.EOL;
			}
			if(j+1!=n) featName += "_";
		}
		if(optionsModel.isRemoveStopwords() && !containsNon && optionsModel.isStopwordsAll()){
			return null;
		}else {
			return featName;
		}
	}

	/**
	 * This method is used when no advanced features are specified. It
	 * checks the static variables of the FeatureSelector, which are, in turn,
	 * set by the OptionsModel of the classifier being trained. The only feature
	 * that is not extracted in this method is sentence length. This method
	 * extracts all the features that have a string value, and sentence length
	 * has a numeric value.
	 * 
	 * @param text -
	 *            The text of the current DataPoint for which features are being
	 *            extracted.
	 * @return - The set of features extracted from this text.
	 */
	public THashMap getFeatureSet(String text, String customPrefix) {

		text = text.trim();
		THashMap dataPointFeatureMap = new THashMap();
		Tokens tokens = null;
		if (optionsModel.isStem()) {
			tokens = Tokenizer.getInstance(this.optionsModel).lemmatizeIncStopword(text.toLowerCase());
		} else {
			tokens = Tokenizer.getInstance(this.optionsModel).tokenizeIncStopword(text.toLowerCase());
		}
		for(Integer n : optionsModel.getNGrams()){
			getNGram(tokens, text, customPrefix, dataPointFeatureMap, n);
		}
		if(optionsModel.isLineLength()){
			dataPointFeatureMap.put(new IndexedFeature(customPrefix+FeatureSelector.LINE_LENGTH, -1, 0), 0.0+tokens.getNumTokens());
		}

		// POS BIGRAMS
		if (optionsModel.isPosBigrams()) {
			// re-introduce punctuation, which is required for better POS-tagging
			// (getPOSBigrams internally does sentence segmentation of the input text)
			String newText = HelperFunctions.replaceSpecialTokensWithPunc(text, optionsModel.getLanguage());
			Tokens posTokens = Tokenizer.getInstance(this.optionsModel).getPOSBigrams(newText);
			Iterator itr = posTokens.getIterator();
			while(itr.hasNext()){
				Token token = (Token)itr.next();
				addFeatToFeatureMap(dataPointFeatureMap, HelperFunctions.replacePuncWithSpecialTokens(customPrefix+token.getIndexToken(), optionsModel.getLanguage()), -1,0);
			}
//			getFeaturesFromPos(posTokens, tokens, customPrefix, dataPointFeatureMap);
		}
		// PUNCTUATION - Extract Punctuation. The last character in the sentence
		// is checked against the inventory of known punctuation marks, in
		// punctStringMap
		if (optionsModel.isPunctuation()) {

			if (text.length() > 0)
			{
				String tmp = text.trim();//.toLowerCase();
				String[] tmplist = tmp.split("[ \t\n\r\f]");

				for(String token:tmplist)
				{
					if (FeatureSelector.punctStringMap.containsKey(token)) 
					{
						addFeatToFeatureMap(dataPointFeatureMap, customPrefix+(String) punctStringMap.get(token), -1,0);
					}

				}
			}
		}
		// CONTAINS NON_STOPWORD - this is true iff any of the words in the text
		// qualifies as a non-stopwords, based on the list of stopwords in the
		// default.stp file in the ROOT directory/etc. See StopwordManager.
		if (optionsModel.isContainsNonStopword()) {
			tokens = Tokenizer.getInstance(this.optionsModel).tokenizeIncStopword(text.toLowerCase());
			Iterator tokenIter = tokens.getIterator();
			boolean containsNonStopword = false;
			while (tokenIter.hasNext()) {
				Token token = (Token) tokenIter.next();
				if (!StopwordManager.getInstance(this.optionsModel).isStopword(token.getIndexToken())) {
					containsNonStopword = true;
					break;
				}
			}
			if(containsNonStopword){
				addFeatToFeatureMap(dataPointFeatureMap, customPrefix + FeatureSelector.CONTAINS_NON_STOPWORD, -1,0);
			}
		}
		return dataPointFeatureMap;   	
	}


	/**
	 * Private utility function that adds or increments frequency count of the given 
	 * feature to the ordinary feature map.
	 * 
	 * @param dataPointFeatureMap
	 * @param feature
	 */
	private void addFeatToFeatureMap(THashMap dataPointFeatureMap, String feature, int index, int length) {
		if (!feature.equals(Token.EMPTY)) {
			IndexedFeature ft = new IndexedFeature(feature, index, length);
			if (dataPointFeatureMap.containsKey(ft)) {
				Double freq = (Double) dataPointFeatureMap.get(ft);
				freq++;
				dataPointFeatureMap.put(ft, freq);
			} else {
				dataPointFeatureMap.put(ft, new Double(1));
			}
		}
	}


	public OptionsModel getOptionsModel() {   
		return this.optionsModel;
	}


	public void setMinCount(int minCount) {
		this.minCount = minCount;
	}

	public int getMinCount() {
		return this.minCount;
	}

	public void setMaxCount(int maxCount) {
		this.maxCount = maxCount;
	}

	public int getMaxCount() {
		return this.maxCount;
	}

	public int countWords(String text) {
		//System.err.println(text);
		return Tokenizer.getInstance(this.optionsModel).tokenizeIncStopword(text).getNumTokens();
	}
	
	public static THashMap getPunct(){
		return punctStringMap;
	}

	public double getLineLength(String text)
	{
		if(this.maxCount ==Integer.MIN_VALUE || this.minCount == Integer.MAX_VALUE )
		{
			return (double)this.countWords(text);
		}
		else
		{
			int wordCount = this.countWords(text);
			double num = (double) (wordCount - this.minCount);
			double denom = (double) (this.maxCount - this.minCount);
			return (double) (num / denom);
		}
	}


	public void addFeatures(THashSet featureSet) {
		Iterator iter = featureSet.iterator();
		while (iter.hasNext()) {
			String feature = iter.next().toString();
			if (this.featureMap.contains(feature)) {
				int count = ((Integer) this.featureMap.get(feature)).intValue();
				count++;
				featureMap.put(feature, new Integer(count));
			} else {
				this.featureMap.put(feature, new Integer(1));
			}
		}
	}


	/**
	 * Feature counts are collected because the user can optionally remove rare
	 * features (features that occur less than THRESH number of times). Those
	 * features that occur less than THRESH number of times are removed from the
	 * featureSet
	 */
	public void removeRareFeatures() {
		Iterator featureIter = this.featureMap.keySet().iterator();
		// get the threshold that was set from the user iterface
		int thresh = this.optionsModel.getFrequencyThreshold();
		while (featureIter.hasNext()) {
			String feature = featureIter.next().toString();
			Object tmp =  featureMap.get(feature);
			int count=Integer.MIN_VALUE;
			if(tmp.getClass()==Integer.class)
				count = ((Integer)tmp).intValue();

			if (count < thresh) {
				featureIter.remove();
			}
		}
	}


	public THashMap getFeatureMap()
	{
		return this.featureMap;
	}


	/**
	 * @param args
	 */
	public static void main(String[] args) {

		//assuming 
		String[] trainStrs = {"A set of tests specifically for Congo fever were still outstanding."
				,"The woman was admitted to the hospital on Saturday after complaining of severe joint pains."
				,"He said it was his opinion that the patient -- a woman -- was suffering from tick bite fever."
				,"A German source said Germany had pledged 153 million marks (66.32 million dollars)."};

		OptionsModel optionsModel = new OptionsModel();
		//set the options in the optionsModel here
		//or use default setting in the optionsModel
		//....

		FeatureSelector featureSelector = new FeatureSelector(optionsModel);

		for(String trainStr:trainStrs) {
			System.err.println(trainStr);


			//step 1: replace punctuations with special tokens
			trainStr = HelperFunctions.replacePuncWithSpecialTokens(trainStr,
					featureSelector.getOptionsModel().getLanguage());
			System.err.println(trainStr);

			//step 2: get the feature set for the text
			THashMap featureMap = featureSelector.getFeatureSet(trainStr);
			System.err.println(featureMap);

			//step 3: prepare for line length feature
			int numTokens = 0;
			numTokens = featureSelector.countWords(trainStr);
			// updating the min/max token count for the feature selector.
			// min/max are used if the user specifies that they want to use
			// sentence length as a feature: sent_length = (X - min)/(max -
			// min)
			if (numTokens > featureSelector.getMaxCount()) {
				featureSelector.setMaxCount(numTokens);
			}
			if (numTokens < featureSelector.getMinCount()) {
				featureSelector.setMinCount(numTokens);
			}

			//step 4: put features into a union feature map
			//prepare for remove rare features
			featureSelector.addFeatures(new THashSet(featureMap.keySet()));    


		}

		//line length feature
		if(optionsModel.isLineLength())
			for(String trainStr:trainStrs) {
				//need to replace punctuations with special tokens
				trainStr = HelperFunctions.replacePuncWithSpecialTokens(trainStr,
						featureSelector.getOptionsModel().getLanguage());
				System.err.println(featureSelector.getLineLength(trainStr));
			}

		//remove rare features if required
		if(optionsModel.isRemoveRareFeatures()) {
			featureSelector.removeRareFeatures();
			System.err.println(featureSelector.getFeatureMap());
		}
	}
}
