/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

package edu.cmu.taghelpertools.tokenizer;

import java.util.Iterator;

import java.util.Set;
import javax.swing.JOptionPane;

import com.aliasi.chunk.Chunk;
import com.aliasi.chunk.Chunking;
import com.aliasi.sentences.MedlineSentenceModel;
import com.aliasi.sentences.SentenceChunker;
import com.aliasi.sentences.SentenceModel;
import com.aliasi.tokenizer.IndoEuropeanTokenizerFactory;
import com.aliasi.tokenizer.TokenizerFactory;

import edu.cmu.taghelpertools.middle_layer.FeatureSelector;
import edu.cmu.taghelpertools.util.HelperFunctions;
import edu.stanford.nlp.tagger.maxent.MaxentTagger;

/**
 * This Singleton class is the single point where features are extracted from
 * the text. This class is used to extract unigrams, bigrams, and POS bigrams
 * from the text. The user can manipulate two extraction parameters: (1)
 * stemming and (2) stopword removal, producing 4 different combinations: (1)
 * stemming, while including stopwords, (2) stemming, while excluding stopwords,
 * (3) not stemming, while including stopwords, and (4) not stemming, while
 * excluding stopwords. There is a method for each. There are also methods for
 * doing these tasks with bigrams. Each method returns a Tokens object. The
 * Tokens object is a list if Token objects. Each Token has a surface string and
 * an index string. The surface string is the exact same string that is
 * originally read in. The index string depends on the options set by the user.
 * If the user is stemming, the index string is the stem, while the surface form
 * is the original string. If the user is removing stopwords, and this word is a
 * stopword, then the index string is the filler "<EMPTY>" and the and the
 * surface form is the original string.
 * 
 * @author Jaime Arguello
 * 
 */
public class EnglishTokenizer extends AbstractTokenizer{

	//    private static final String BOL = "BOL";
	//    private static final String EOL = "EOL";
	private static MaxentTagger tagger;

	// sentence segmentation stuff
	static final TokenizerFactory TOKENIZER_FACTORY = new IndoEuropeanTokenizerFactory();
	static final SentenceModel SENTENCE_MODEL = new MedlineSentenceModel();
	static final SentenceChunker SENTENCE_CHUNKER = new SentenceChunker(TOKENIZER_FACTORY, SENTENCE_MODEL);
	public static String STOPWORD_FILENAME = Tokenizer.STPS[Tokenizer.LANG_ENGLISH];  


	static {
		try {
			tagger = new MaxentTagger("toolkits/maxent/left3words-wsj-0-18.tagger");
			// check if we are to use a custom stoplist
			// 
			// this should be only a file name with the file being present in
			// the etc/ directory of TagHelperTools2
			String customStop = System.getProperty("tht.customstop");
			if (customStop != null)
				STOPWORD_FILENAME = "etc/" + customStop;
		} catch (Exception e) {
			e.printStackTrace();
			JOptionPane.showMessageDialog(null, "Could not find MaxentTagger files", "ERROR", JOptionPane.ERROR_MESSAGE);
			System.exit(0);
		}
	}

	//    private static Tokenizer instance = new Tokenizer();

	//    private int lineLength;

	protected EnglishTokenizer() {
	}

	//    public static Tokenizer getInstance() {
	//        return instance;
	//    }

	//    public int getLineLength() {
	//        return lineLength;
	//    }

	/**
	 * Extracts set of unigrams, while stemming and removing stopwords.
	 * 
	 * @param line -
	 *            input String to be tokenized.
	 * @return - tokens
	 */
	public Tokens lemmatizeExcStopword(String line) {
		Tokens tokens = new Tokens();
		if (line == null)
			return tokens;
		else {
			if (line.trim().length() == 0) {
				return tokens;
			}
		}
		line = line.trim().toLowerCase();
		//String[] split = line.split("([\\W|_&&[^'-]])");
		String[] split = line.split("([\\W|_&&[^'-<>]])");
		int offset = 0;
		for (int i = 0; i < split.length; i++) {
			String term = split[i];
			if (term.length() > 0) {
				boolean isPreStemmingStop = false;
				if (StopwordManager.getInstance(EnglishTokenizer.STOPWORD_FILENAME).isStopword(term)) {
					isPreStemmingStop = true;
				}
				String stem = EnglishStemmer.stem(term);
				if (StopwordManager.getInstance(EnglishTokenizer.STOPWORD_FILENAME).isStopword(stem) || isPreStemmingStop) {
					tokens.addToken(Token.EMPTY, term, offset);
				} else {
					tokens.addToken(stem, term, offset);
					//                    if(fsm!=null)
					//                    	fsm.addOrdFeature(stem, i, 1, false);
					//                    	//fsm.addOrdFeature("THF_UStemRmstopw_"+stem, i, 1, false);                      
				}
				offset += term.length()+1;					
			}
		}
		return tokens;
	}

	/**
	 * Extracts set of unigrams, while stemming. Stopwords are not removed.
	 * 
	 * @param line -
	 *            input String to be tokenized.
	 * @return - tokens
	 */
	public Tokens lemmatizeIncStopword(String line) {
		Tokens tokens = new Tokens();
		if (line == null)
			return tokens;
		else {
			if (line.trim().length() == 0) {
				return tokens;
			}
		}
		line = line.trim().toLowerCase();
		//String[] split = line.split("([\\W|_&&[^'-]])");
		String[] split = line.split("([\\W|_&&[^'-<>]])");
		int offset = 0;
		for (int i = 0; i < split.length; i++) {
			String term = split[i];
			if (term.length() > 0) {
				String stem = EnglishStemmer.stem(term);
				tokens.addToken(stem, term, offset);
				offset += term.length()+1;	
			}
		}
		return tokens;
	}

	/**
	 * Extracts set of unigrams. Does not stem or remove stopwords.
	 * 
	 * @param line -
	 *            input String to be tokenized.
	 * @return - tokens
	 */
	public Tokens tokenizeIncStopword(String line) {
		Tokens tokens = new Tokens();
		if (line == null || (line.trim().length() == 0))
			return tokens;
		line = line.trim().toLowerCase();
		//String[] split = line.split("([\\W|_&&[^'-]])");
		String[] split = line.split("([\\W|_&&[^'-<>]])");
		int offset = 0;
		for (int i = 0; i < split.length; i++) {
			String term = split[i];
			if (term.length() > 0) {
				tokens.addToken(term, term, offset);
				//                if(fsm!=null)
				//                	//THF = TagHelper Features; UNostemStop = unigram + nostem + with stopword
				//                	//fsm.addOrdFeature("THF_UNostemNormstopw_"+term, i, 1, false);
				//                	fsm.addOrdFeature(term, i, 1, false);
				offset += term.length()+1;	
			}
		}
		return tokens;
	}

	/**
	 * Extracts set of unigrams. No stemming is done. Stopwords are removed.
	 * 
	 * @param line -
	 *            input String to be tokenized.
	 * @return - tokens
	 */
	public Tokens tokenizeExcStopword(String line) {
		Tokens tokens = new Tokens();
		if (line == null || (line.trim().length() == 0))
			return tokens;
		line = line.trim().toLowerCase();
		//String[] split = line.split("([\\W|_&&[^'-]])");
		String[] split = line.split("([\\W|_&&[^'-<>]])");
		int offset = 0;
		for (int i = 0; i < split.length; i++) {
			String term = split[i];
			if (term.length() > 0) {
				if (!StopwordManager.getInstance(EnglishTokenizer.STOPWORD_FILENAME).isStopword(term)) {
					tokens.addToken(term, term, offset);
					//                    if(fsm!=null)
					//                    	//fsm.addOrdFeature("THF_UNostemRmstopw_"+term, i, 1, false);                      
					//                    	fsm.addOrdFeature(term, i, 1, false);
				} else {
					tokens.addToken(Token.EMPTY, term, offset);
				}
				offset += term.length()+1;	
			}
		}
		return tokens;
	}

	/**
	 * Checks to see the input character is an English character or a Digit.
	 * This will only work for English!
	 * 
	 * @param ch
	 * @return true if the input is an English letter or a digit and false
	 *         otherwise.
	 */
	static boolean isLetterOrDigit(char ch) {
		return ((ch >= '0' && ch <= '9') || (ch >= 'A' && ch <= 'Z') || (ch >= 'a' && ch <= 'z'));
	}


	/**
	 * Removes non-letter characters from the input text. This code is
	 * English-specific, it will cause unwanted results in other languages
	 * (e.g., assume everything is a non-valid char). For TagHelper to work in
	 * languages other than English, this method needs to be customized. As it
	 * stands right now, all letters that are outside the range of the English
	 * character set are removed.
	 * 
	 * @param term
	 *            to be cleaned
	 * @return - cleaned term
	 */
	public String trimNonLetterChars(String term) {

		// this code is English-specific, so I'm removing it -- Jaime
		int start = 0;
		int end = term.length() - 1;
		char[] chars = term.toCharArray();
		for (int i = 0; i < chars.length; i++) {
			if (!EnglishTokenizer.isLetterOrDigit(chars[i])) {
				start++;
			} else {
				break;
			}
		}
		for (int i = term.length() - 1; i >= 0; i--) {
			if (!EnglishTokenizer.isLetterOrDigit(chars[i])) {
				end--;
			} else {
				break;
			}
		}
		String newTerm = "";
		for (int i = start; i <= end; i++) {
			newTerm += ("" + chars[i]);
		}
		return newTerm;
	}

	/**
	 * Extracts POS bigrams from input text using the Stanford POS tagger.
	 * 
	 * @param line
	 *            to be processed
	 * @return - Tokens with POS bigrams
	 */
	public Tokens getPOSBigrams(String line)
	{
		return getPOSBigrams(line, true);
	}

	public Tokens tokenizePOS(String line){
		Tokens tokens = new Tokens();
		int length_limit = 200; // original -->300
		String taggedString = "";
		Chunking chunking = SENTENCE_CHUNKER.chunk(line.toCharArray(), 0, line.length());
		Set sentences = chunking.chunkSet();
		String slice = chunking.charSequence().toString();
		AbstractTokenizer.taggedResult = "";
		for (Iterator it = sentences.iterator(); it.hasNext();) {
			Chunk sentenceChunk = (Chunk) it.next();
			int starti = sentenceChunk.start();
			int endi = sentenceChunk.end();
			String sentence = slice.substring(starti, endi);
			String[] surface = sentence.split("\\W+");
			try {            
				if (surface.length > length_limit) {
					//fixing the issue of overlengthy sentences
					String[] tmpTokens = sentence.split("\\W+");
					int num_buckets = (int)Math.ceil((double)tmpTokens.length / (double)length_limit );
					StringBuffer sb = new StringBuffer();
					for(int i=0; i< num_buckets; i++)
					{

						String tmpResult = "";
						if(i != num_buckets-1)
							tmpResult = tagger.tagString(getRegularizedString(tmpTokens, i*length_limit, (i+1)*length_limit-1)); 
						else
							tmpResult = tagger.tagString(getRegularizedString(tmpTokens, i*length_limit, tmpTokens.length-1)); 

						sb.append(tmpResult.trim()).append(" ");
					}
					taggedString = sb.toString().trim();
				}
				else{
					taggedString = tagger.tagString(sentence);	           
				}
			} 
			catch (Exception e) {
				POSerror(sentence);
				continue;
			} 
			AbstractTokenizer.taggedResult = AbstractTokenizer.taggedResult+" "+taggedString;
			if (taggedString == null || taggedString.length() == 0) {
				POSerror(sentence);
				continue;
			}
			String[] split = taggedString.split(" ");
			for (int i = 0; i < split.length; i++) {
				String pos = split[i].split("\\/")[1];
				Token t = new Token(surface[i], surface[i], i);
				t.setPos(pos);
				tokens.addToken(t);
			}
		}
		return tokens;
	}

	public void POSerror(String sentence){
		System.out.println("WARNING! COULD NOT PARSE: " + sentence);
	}

	public static String[] getSentenceArray(String line){
		Chunking chunking = SENTENCE_CHUNKER.chunk(line.toCharArray(), 0, line.length());
		Set sentences = chunking.chunkSet();
		String slice = chunking.charSequence().toString();
		String[] sentencesOut = new String[sentences.size()];
		int i = 0;
		for (Iterator it = sentences.iterator(); it.hasNext();) {
			Chunk sentenceChunk = (Chunk) it.next();
			int starti = sentenceChunk.start();
			int endi = sentenceChunk.end();
			String sentence = slice.substring(starti, endi);
			sentencesOut[i++] = sentence;
		}
		return sentencesOut;
	}

	/**
	 * Extracts POS bigrams from input text using the Stanford POS tagger. 
	 * The issue of not able to handle overlength strings has been addressed by 
	 * Hao-Chuan Wang in Dec/09/2006 by modifying this Method.
	 * 
	 * @param line
	 *            to be processed
	 * @param fix_overlength by Hao-Chuan Wang, December 2006
	 * @return - Tokens with POS bigrams
	 */
	public Tokens getPOSBigrams(String line, boolean fix_overlength) {
		Tokens tokens = new Tokens();

//		System.out.println("ET389: " + line + "\n");
		int length_limit = 200; // original -->300
		String taggedString = "";
		Chunking chunking = SENTENCE_CHUNKER.chunk(line.toCharArray(), 0, line.length());
		Set sentences = chunking.chunkSet();
		String slice = chunking.charSequence().toString();
		AbstractTokenizer.taggedResult = "";
		int offset = 0;
		for (Iterator it = sentences.iterator(); it.hasNext();) {
			Chunk sentenceChunk = (Chunk) it.next();
			int starti = sentenceChunk.start();
			int endi = sentenceChunk.end();
			offset = starti;
			String sentence = slice.substring(starti, endi);
			try {            
				if (sentence.split("\\W+").length > length_limit) {
					if(!fix_overlength)
					{
						System.out.println("SENTENCE TOO LONG! COULD NOT PARSE: " + sentence);
						continue;
					}
					else
					{
						//fixing the issue of overlengthy sentences
						String[] tmpTokens = sentence.split("\\W+");
						int num_buckets = (int)Math.ceil((double)tmpTokens.length / (double)length_limit );
						StringBuffer sb = new StringBuffer();
						for(int i=0; i< num_buckets; i++)
						{

							String tmpResult = "";
							if(i != num_buckets-1)
								tmpResult = tagger.tagString(getRegularizedString(tmpTokens, i*length_limit, (i+1)*length_limit-1)); 
							else
								tmpResult = tagger.tagString(getRegularizedString(tmpTokens, i*length_limit, tmpTokens.length-1)); 

							sb.append(tmpResult.trim()).append(" ");
						}
						taggedString = sb.toString().trim();
						//System.out.println(taggedString);
					}
				}
				else{
					taggedString = tagger.tagString(sentence);	           
				}
			} 
			catch (Exception e) {
				System.out.println("WARNING! COULD NOT PARSE: " + sentence);
				continue;
			} 
			//System.out.println(taggedString);
			AbstractTokenizer.taggedResult = AbstractTokenizer.taggedResult+" "+taggedString;
			if (taggedString == null) {
				System.out.println("WARNING! COULD NOT PARSE: " + sentence);
				continue;
			} else if (taggedString.length() == 0) {
				System.out.println("WARNING! COULD NOT PARSE: " + sentence);
				continue;
			}
			String[] split = taggedString.split(" ");
			String prevPos = EnglishTokenizer.BOL;
			for (int i = 0; i < split.length; i++) {
				// System.out.println(split[i]);
				String pos = split[i].split("\\/")[1];
				String bigram = prevPos + "_" + pos;
				String surf = HelperFunctions.replacePuncWithSpecialTokens(split[i].split("\\/")[0], Tokenizer.LANGS[Tokenizer.LANG_ENGLISH]);
				if(i>0){
					String prevSurf = HelperFunctions.replacePuncWithSpecialTokens(split[i-1].split("\\/")[0], Tokenizer.LANGS[Tokenizer.LANG_ENGLISH]);
					tokens.addToken(bigram, prevSurf+"_"+surf, offset);
				}else{
					tokens.addToken(bigram, surf, offset);					
				}
				//                if(fsm!=null)
				//                	fsm.addOrdFeature(bigram, i, 2, true);                    
				//                	//fsm.addOrdFeature("THF_POS_"+bigram, i, 2, true);
				prevPos = pos;
				if (i == split.length - 1) {
					bigram = prevPos + "_" + EnglishTokenizer.EOL;
					tokens.addToken(bigram, surf, offset);
					offset += surf.length()+1;
					//                    if(fsm!=null)
					//                    	fsm.addOrdFeature(bigram, i, 2, true);    
					//                    	//fsm.addOrdFeature("THF_POS_"+bigram, i, 2, true);
				}
			}
		}
		return tokens;
	}

	/**
	 * Generating regularized strings by putting a space between each string token
	 * @param tokens
	 * @param startIndex
	 * @param endIndex
	 * @return -Regularized String
	 */
	private static String getRegularizedString(String[] tokens, int startIndex, int endIndex)
	{
		StringBuffer sb = new StringBuffer();

		for(int i=0; i<tokens.length; i++)
		{
			if(i>= startIndex && i<=endIndex)
				sb.append(tokens[i]).append(" ");
		}

		return sb.toString();
	}

	public static void main(String args[]) {
		try{

			String test = "I am wondering what the issue is here..";
			System.out.println(test.split("\\W+").length);
			MaxentTagger tg = new MaxentTagger("toolkits/maxent/left3words-wsj-0-18.tagger");
			tg.tagString(test);
		}catch(Exception e){
			e.printStackTrace();
		}
	}
}
