/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

package edu.cmu.taghelpertools.tokenizer;

import java.util.ArrayList;
import java.util.Iterator;

/**
 * A list of tokens. Used as output to all methods in Tokenizer singleton class.
 * Each Token has a surface string and an index string. The surface string is
 * the exact same string that is originally read in. The index string depends on
 * the options set by the user. If the user is stemming, the index string is the
 * stem, while the surface form is the original string. If the user is removing
 * stopwords, and this word is a stopword, then the index string is the filler "<EMPTY>"
 * and the surface form is the original string.
 * 
 * @author Jaime Arguello
 * 
 */
public class Tokens {

	private ArrayList tokens;

	public Tokens() {
		tokens = new ArrayList();
	}
	
	/**
	 * Adds a new token to the token list.
	 * 
	 * @param indexToken
	 * @param surfaceToken
	 */
	public void addToken(String indexToken, String surfaceToken, int index) {
		tokens.add(new Token(indexToken, surfaceToken, index));
	}

	/**
	 * Adds TokenNode to the token list.
	 * 
	 * @param tokenNode
	 */
	public void addToken(Token tokenNode) {
		tokens.add(tokenNode);
	}

	/**
	 * Adds all tokens to the token list.
	 * 
	 * @param tokens
	 */
	public void putAll(Tokens tokens) {
		Iterator iter = tokens.getIterator();
		while (iter.hasNext()) {
			Token tNode = (Token) iter.next();
			this.addToken(tNode);
		}
	}

	/**
	 * Adds the token to the token list *mag* number of times.
	 * 
	 * @param surfaceToken
	 * @param indexToken
	 * @param mag
	 */
	public void addToken(String surfaceToken, String indexToken, int index, int mag) {
		for (int i = 0; i < mag; i++) {
			tokens.add(new Token(surfaceToken, indexToken, index));
		}
	}

	/**
	 * Returns an iterator for accessing tokens.
	 * 
	 * @return
	 */
	public Iterator getIterator() {
		return tokens.iterator();
	}

	/**
	 * Returns the number of tokens in the token list. This method comes in
	 * handy when getting the length of a sentence (in # of words).
	 * 
	 * @return
	 */
	public int getNumTokens() {
		return tokens.size();
	}
	
	@Override
	public String toString(){
		Iterator<Token> it = getIterator();
		StringBuilder sb = new StringBuilder();
		while(it.hasNext()){
			sb.append(it.next().getSurfaceToken());
			if(it.hasNext()) sb.append(" ");
		}
		return sb.toString();
	}
	

}