package plugin.simple.learning;

import java.awt.Component;

import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;

import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;
import java.util.TreeMap;
import java.util.TreeSet;

import javax.swing.JLabel;

import org.w3c.dom.Element;

import weka.classifiers.AbstractClassifier;
import weka.classifiers.Classifier;
import weka.classifiers.Evaluation;
import weka.classifiers.bayes.NaiveBayes;
import weka.classifiers.functions.SMO;
import weka.core.Instance;
import weka.core.Instances;
import weka.gui.GenericObjectEditor;
import weka.gui.PropertyPanel;

import edu.cmu.side.SimpleWorkbench;
import edu.cmu.side.dataitem.TrainingResultInterface;
import edu.cmu.side.simple.LearningPlugin;
import edu.cmu.side.simple.SimpleDocumentList;
import edu.cmu.side.simple.SimpleTrainingResult;
import edu.cmu.side.simple.feature.Feature;
import edu.cmu.side.simple.feature.FeatureTable;

public class WekaLearning extends LearningPlugin{
	private static final long serialVersionUID = 4897708453492254724L;

	private transient PropertyPanel propertyPanel;
	private transient GenericObjectEditor genericObjectEditor;

	private Classifier classifier;

	public void setClassifier(Classifier c){
		classifier = c;
	}
	
	public Classifier getClassifier(){
		return classifier;
	}
	
	public WekaLearning(){
		genericObjectEditor = new GenericObjectEditor();
		genericObjectEditor.setClassType(weka.classifiers.Classifier.class);
		genericObjectEditor.setValue(new weka.classifiers.bayes.NaiveBayes());
		propertyPanel = new PropertyPanel(genericObjectEditor);
		halt = false;
	}
	@Override
	public String getOutputName() {
		return "weka";
	}

	@Override
	protected Component getConfigurationUIForSubclass() {
		return propertyPanel;
	}

	@Override
	public String toString(){
		return getOutputName();
	}

	@Override
	public TrainingResultInterface train(FeatureTable featureTable,
			String desiredName, Map<String, String> evaluationSettings, Map<Integer, Integer> foldsMap, JLabel progressLabel) throws Exception{
		progressLabel.setText("Converting data to Weka format");
		featureTable.generateFastVector();
		progressLabel.setText("Training classifier");
		classifier = (Classifier)genericObjectEditor.getValue();
		double time1 = System.currentTimeMillis();
		classifier.buildClassifier(featureTable.getInstances());			
		double time2 = System.currentTimeMillis();
		progressLabel.setText((time2-time1) + " ms to get instances and train.");
		SimpleTrainingResult result = new SimpleTrainingResult(this, desiredName, featureTable);
		result.addEvaluation(evaluate(result, classifier, evaluationSettings, foldsMap, progressLabel));	
		progressLabel.setText("");
		return result;
	}

	public Map<String, String> evaluate(SimpleTrainingResult result, Classifier classifier, Map<String, String> settings, Map<Integer, Integer> foldsMap, JLabel progressLabel) throws Exception{
		FeatureTable featureTable = result.getFeatureTable();
		Map<String, String> evaluationStrings = new TreeMap<String, String>();
		String[] labelArray = featureTable.getDocumentList().getLabelArray();
		try{
			StringBuilder predictions = new StringBuilder();
			Evaluation eval = new Evaluation(featureTable.getInstances());
			DecimalFormat print = new DecimalFormat("#.##");
			Map<Integer, String> predicted = new TreeMap<Integer, String>();
			ArrayList<Integer> indexMap = new ArrayList<Integer>();
			ArrayList<Double> pred = new ArrayList<Double>();
			ArrayList<Integer> folds = new ArrayList<Integer>();
			if(settings.containsKey("test-set")){
				System.out.println("WL99 evaluating with test set.");
				evaluateWithTestSet(result, classifier, settings, eval,indexMap, pred, progressLabel);
			}else {
				evaluateWithCrossValidation(featureTable, classifier, foldsMap, progressLabel, eval, print, indexMap, pred, folds);
				if(halt){
					halt = false;
					return null;
				}
			}
			finalizeEvaluation(featureTable, labelArray, predictions, predicted, indexMap, pred);
			getModelSpecificDetails(featureTable, classifier);
			evaluationStrings.put("cv-fold-predictions", predictions.toString());
			evaluationStrings.put("summary", eval.toSummaryString());
			evaluationStrings.put("model", classifier.toString());;
			if(!featureTable.getClassValueType().equals(Feature.Type.NUMERIC)){
				evaluationStrings.put("confusion matrix", eval.toMatrixString());				
			}
		}catch(Exception e){
			e.printStackTrace();
		}
		return evaluationStrings;
	}
	private void finalizeEvaluation(FeatureTable featureTable,
			String[] labelArray, StringBuilder predictions,
			Map<Integer, String> predicted, ArrayList<Integer> indexMap,
			ArrayList<Double> pred) {
		int evalIndex = 0;
		switch(featureTable.getClassValueType()){
		case NOMINAL:
		case BOOLEAN:
			for(int ind : indexMap){
				predicted.put(ind, labelArray[((Double)(pred.get(evalIndex++))).intValue()]);
			}
			break;
		case NUMERIC:
			for(int ind : indexMap){
				predicted.put(ind, ""+pred.get(evalIndex++));
			}
			break;
		}
		for(int i = 0; i < predicted.keySet().size(); i++){
			predictions.append(predicted.get(i) + "\n");
		}
	}
	private void getModelSpecificDetails(FeatureTable featureTable,
			Classifier classifier) {
		if(classifier instanceof SMO){
			String[] model = classifier.toString().split("\n");
			Map<Feature, Comparable> weights = new HashMap<Feature, Comparable>();
			for(String m : model){
				int index = m.indexOf(" * (normalized) ");
				if(index != -1){
					String featureName = m.substring(index+16);
					String[] start = m.substring(0, index).split("\\s+");
					Double weight = Double.parseDouble(start[start.length-1]);
					for(Feature f : featureTable.getFeatureSet()){
						if(f.getFeatureName().equals(featureName)){
							weights.put(f, weight);
						}
					}
				}
			}

			int w = 1;
			while(featureTable.getEvaluations().containsKey("weights"+w)) w++;
			featureTable.addEvaluation("weights"+w, weights);
		}
		if(classifier instanceof NaiveBayes){

		}
	}
	
	@Override
	public void addEvaluationColumns(FeatureTable ft){
		getModelSpecificDetails(ft, classifier);
	}
	private void evaluateWithCrossValidation(FeatureTable featureTable,
			Classifier classifier, Map<Integer, Integer> foldsMap,
			JLabel progressLabel, Evaluation eval, DecimalFormat print,
			ArrayList<Integer> indexMap, ArrayList<Double> pred,
			ArrayList<Integer> folds) throws Exception {
		for(int i : foldsMap.keySet()) if(!folds.contains(foldsMap.get(i))) folds.add(foldsMap.get(i));
		for(int i : folds){
			if(halt){
				return;
			}
//			Significance tests only
//			eval = new Evaluation(featureTable.getInstances());
			
			double time1 = System.currentTimeMillis();		
			Instances train = featureTable.getInstances(foldsMap, i, true);
			eval.setPriors(train);
			Classifier copiedClassifier = AbstractClassifier.makeCopy(classifier);
			copiedClassifier.buildClassifier(train);
			Instances test = featureTable.getInstances(foldsMap, i, false);
			double[] foldPred = eval.evaluateModel(copiedClassifier, test);
			for(double d : foldPred) pred.add(d);
			Map<Integer, Integer> foldMap = featureTable.foldIndexToIndex(foldsMap, i);
			for(int j = 0; j < foldMap.keySet().size(); j++){
				indexMap.add(foldMap.get(j));
			}
//			Significance tests only
//			System.out.println(eval.kappa());
			double time2 = System.currentTimeMillis();
			progressLabel.setText("Fold " + i + " done (" + (time2-time1) + "ms, " + print.format(eval.pctCorrect()) + "% so far)");
		}
	}
	private void evaluateWithTestSet(SimpleTrainingResult result,
			Classifier classifier, Map<String, String> settings,
			Evaluation eval, ArrayList<Integer> indexMap, ArrayList<Double> pred, JLabel progressLabel)
			throws Exception {
		FeatureTable featureTable = result.getFeatureTable();
		TreeSet<String> filenames = new TreeSet<String>();
		filenames.add(settings.get("test-set"));
		System.out.println("Evaluating with test set.");
		progressLabel.setText("Loading test set");
		SimpleDocumentList testSet = new SimpleDocumentList(filenames, featureTable.getDocumentList().getCurrentAnnotation(), featureTable.getDocumentList().getTextColumn());
		progressLabel.setText("Building test set table");
		FeatureTable newTable = new FeatureTable(featureTable.getExtractors(), testSet, 0);
		progressLabel.setText("Reconciling feature tables");
		newTable = FeatureTable.reconcileFeatures(featureTable, newTable);
		progressLabel.setText("Evaluating model");
		newTable.generateFastVector();
		double[] testSetPredictions = eval.evaluateModel(classifier, newTable.getInstances());
		result.setEvaluationTable(newTable);
		for(int i = 0; i < testSetPredictions.length; i++){
			pred.add(testSetPredictions[i]);
			indexMap.add(i);
		}
	}

	public void predict(String name, FeatureTable featureExtractedNewData){
		featureExtractedNewData.generateFastVector();
		Instances inst = featureExtractedNewData.getInstances();

		double[] preds = new double[inst.numInstances()];
		try{
			for(int i = 0; i < inst.numInstances(); i++){
				Instance a = inst.instance(i);
				double[] distri = classifier.distributionForInstance(a);
				preds[i] = classifier.classifyInstance(inst.instance(i));			
			}
		}catch(Exception e){
			e.printStackTrace();
		}
		ArrayList<String> annots = new ArrayList<String>();
		switch(featureExtractedNewData.getClassValueType()){
		case BOOLEAN:
		case NOMINAL:
			String[] labels = featureExtractedNewData.getDocumentList().getLabelArray();
			for(int i = 0; i < preds.length; i++){
				annots.add(labels[(int)preds[i]]);
			}
			break;
		case NUMERIC:
			for(int i = 0; i < preds.length; i++){
				annots.add(""+preds[i]);
			}
			break;
		}
		featureExtractedNewData.getDocumentList().addAnnotation(name, annots);
	}
	@Override
	public void fromFile(double uniqueID) {
		fromFile(SimpleWorkbench.savedFolder+"/model_"+uniqueID+".ser");
	}
	
	public void fromFile(String filename) {
		try{
			ObjectInputStream in = new ObjectInputStream(new FileInputStream(filename));
			System.out.println(filename);
			genericObjectEditor.setValue((Classifier)in.readObject());
			in.close();
		}catch(Exception e){
			e.printStackTrace();
		}
	}
	
	@Override
	public void configureFromFile(String f){
		fromFile(f);
	}

	@Override
	public void toFile(double uniqueID) {
		try{
			ObjectOutputStream out = new ObjectOutputStream(new FileOutputStream(SimpleWorkbench.savedFolder+"/model_"+uniqueID+".ser"));
			out.writeObject(classifier);
			out.close();
		}catch(Exception e){
			e.printStackTrace();
		}
	}
}
